/*
 *  membandwidth.c
 *
 *  Written by
 *   Andreas Dehmel <zarquon@t-online.de>
 *
 *  Simple ANSI-C program to determine a computer's memory bandwidth
 *  in cache and main memory. It performs several operations on memory
 *  blocks (1k and 8M by default) for a certain number of seconds (10
 *  by default) to achieve this. Operations are memcpy, memset and
 *  an blocksum (which essentially should reflect the read-only speed).
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>


#define BLOCKSUM_UNROLL


typedef struct memtest_env_s {
  int *src, *dest;
  size_t blocksize;
} memtest_env_t;

static int memtest_env_init(memtest_env_t *me, size_t bsize)
{
  me->blocksize = (bsize + sizeof(int)-1) & ~(sizeof(int)-1);
  me->src = (int*)malloc(me->blocksize);
  me->dest = (int*)malloc(me->blocksize);
  if ((me->src != 0) && (me->dest != 0))
  {
    memset(me->src, 0, sizeof(me->blocksize));
    memset(me->dest, 0, sizeof(me->blocksize));
    return 0;
  }
  return -1;
}

static void memtest_env_exit(memtest_env_t *me)
{
  if (me->src != 0)
  {
    free(me->src);
    me->src = 0;
  }
  if (me->dest != 0)
  {
    free(me->dest);
    me->dest = 0;
  }
  me->blocksize = 0;
}

static double test_bandwidth_function(void (*f)(memtest_env_t*), const char *name, memtest_env_t *me, int seconds)
{
  time_t curTime, startTime = time(0);
  unsigned long number, counter;
  double bandwidth;

  printf("%s...", name);
  fflush(stdout);
  do
  {
    curTime = time(0);
  }
  while (curTime == startTime);

  startTime = curTime;
  number = 0; counter = 0;
  do
  {
    f(me);
    number++;
    counter += me->blocksize;
    if (counter >= 1048576)
    {
      curTime = time(0);
      counter = 0;
    }
  }
  while ((int)(curTime - startTime) < seconds);

  bandwidth = (((double)number)*me->blocksize) / (1048576.0*seconds);

  printf(" %f MB/s\n", bandwidth);

  return bandwidth;
}

static void bandwidth_test_memcpy(memtest_env_t *me)
{
  memcpy(me->dest, me->src, me->blocksize);
}

static void bandwidth_test_memset(memtest_env_t *me)
{
  memset(me->dest, 1, me->blocksize);
}

#ifdef BLOCKSUM_UNROLL
static void bandwidth_test_blocksum8(memtest_env_t *me)
{
  size_t i = me->blocksize/sizeof(int);
  const int *src = me->src;
  int sum = 0;

  while (i >= 8)
  {
    sum += src[0];
    sum += src[1];
    sum += src[2];
    sum += src[3];
    sum += src[4];
    sum += src[5];
    sum += src[6];
    sum += src[7];
    i -= 8;
    src += 8;
  }
  if (i >= 4)
  {
    sum += src[0];
    sum += src[1];
    sum += src[2];
    sum += src[3];
    i -= 4;
    src += 4;
  }
  if (i >= 2)
  {
    sum += src[0];
    sum += src[1];
    i -= 2;
    src += 2;
  }
  if (i >= 1)
    sum += src[0];

  (me->dest)[0] = sum;
}
#else
static void bandwidth_test_blocksum(memtest_env_t *me)
{
  size_t i, num = me->blocksize/sizeof(int);
  int sum = 0;
  for (i=0; i<num; i++)
    sum += (me->src)[i];
  (me->dest)[0] = sum;
}
#endif

static double test_bandwidth(size_t bsize, int seconds)
{
  memtest_env_t memenv;
  double bandwidth = -1.0;

  if (memtest_env_init(&memenv, bsize) == 0)
  {
    printf("Block size %d [%d seconds]:\n", memenv.blocksize, seconds);
    bandwidth  = test_bandwidth_function(bandwidth_test_memcpy, "memcpy", &memenv, seconds);
    bandwidth += test_bandwidth_function(bandwidth_test_memset, "memset", &memenv, seconds);
#ifdef BLOCKSUM_UNROLL
    bandwidth += test_bandwidth_function(bandwidth_test_blocksum8, "blocksum8", &memenv, seconds);
#else
    bandwidth += test_bandwidth_function(bandwidth_test_blocksum, "blocksum", &memenv, seconds);
#endif
  }
  memtest_env_exit(&memenv);

  return bandwidth;
}


int main(int argc, const char *argv[])
{
  size_t bsCache = 1024, bsMemory = 8388608;
  double bwCache, bwMemory;
  int i=1, seconds = 10;

  while (i < argc)
  {
    if (argv[i][0] == '-')
    {
      switch (argv[i][1])
      {
        case 'c':
          bsCache = (size_t)atoi(argv[++i]);
          break;
        case 'm':
          bsMemory = (size_t)atoi(argv[++i]);
          break;
        case 's':
          seconds = atoi(argv[++i]);
          break;
        default:
          fprintf(stderr, "Unknown switch %s\n", argv[i]);
          break;
      }
    }
    else
      fprintf(stderr, "Unknown parameter %s\n", argv[i]);

    i++;
  }

  bwCache  = test_bandwidth(bsCache, seconds);
  bwMemory = test_bandwidth(bsMemory, seconds);
  printf("Average cache to main memory factor: %f\n", bwCache / bwMemory);

  return 0;
}
